import math
import pandas as pd

from typing import List, Callable


def get_column_type(df: pd.DataFrame, column: str) -> str:
    unique_values = df[column].astype(str).unique()
    if len(unique_values) <= 1:
        return "ignore"

    dtype = df[column].dtype.name
    if dtype.startswith("int") or dtype.startswith("float"):
        column_type = "continuous"
    else:
        column_type = "categorical"
    return column_type


def get_column_types(df: pd.DataFrame) -> pd.DataFrame:
    col_types = {}
    for column in df.columns:
        col_type = get_column_type(df, column)
        if col_type not in col_types:
            col_types[col_type] = []
        col_types[col_type].append(column)

    cardinality = {}
    for col in col_types["categorical"]:
        cardinality[col] = len(df[col].astype(str).unique())
    return col_types, cardinality


def z_score_scale(
    df: pd.DataFrame,
    column: str,
    clip: bool = False,
    lower: int = None,
    higher: int = None,
) -> pd.DataFrame:
    series = df[column]
    if clip:
        series = series.clip(lower=lower, higher=higher)
    mean = series.mean()
    std = series.std()
    transformed = (series - mean) / std
    df[f"{column}: zscore"] = transformed
    return df


def onehot_encode(df: pd.DataFrame, column: str) -> pd.DataFrame:
    series = df[column].astype(str)
    unique_categories = [
        category for category in sorted(series.unique()) if category not in ["nan"]
    ]
    records = series.map(
        lambda category: {
            f"{column}: {unique_category}": int(unique_category == category)
            for unique_category in unique_categories
        }
    ).tolist()
    new_df = pd.DataFrame.from_records(records)
    return concat_dfs(df, new_df)


def lambda_onehot_encode(df: pd.DataFrame, column: str, fn: Callable) -> pd.DataFrame:
    series = df[column].astype(float)
    max_value = series.max()
    max_lambda_value = fn(max_value)
    df[column] = series.map(lambda value: min(fn(value), max_lambda_value))
    return onehot_encode(df, column)


def sqrt_onehot_encode(df: pd.DataFrame, column: str) -> pd.DataFrame:
    return lambda_onehot_encode(df, column, fn=lambda x: int(x**0.5))


def log_onehot_encode(df: pd.DataFrame, column: str) -> pd.DataFrame:
    return lambda_onehot_encode(df, column, fn=lambda x: int(math.log(x)))


def multihot_encode(df: pd.DataFrame, column: str) -> pd.DataFrame:
    all_categories = set()
    series = df[column]
    series.map(
        lambda categories: [
            all_categories.add(categories)
            for categories in (categories if isinstance(categories, list) else [])
        ]
    )
    all_categories = list(sorted(all_categories))
    records = series.map(
        lambda categories: {
            f"{column}: {unique_category}": unique_category
            in (categories if isinstance(categories, list) else [])
            for unique_category in all_categories
        }
    ).tolist()
    new_df = pd.DataFrame.from_records(records)
    return concat_dfs(df, new_df)


def json_norm_column(df: pd.DataFrame, column: str) -> pd.DataFrame:
    if column not in df.columns:
        return df
    json_df = pd.json_normalize(df[column])
    json_df = json_df[sorted(json_df.columns)]
    json_df = json_df.fillna(0)
    json_df.columns = [f"{column}: {col}" for col in json_df.columns]
    return concat_dfs(df, json_df)


def concat_dfs(*dfs: List[pd.DataFrame], axis: int = 1) -> pd.DataFrame:
    new_dfs = []
    for df in dfs:
        try:
            df = df.reset_index()
        except:
            pass
        new_dfs.append(df)
    concat_df = pd.concat(new_dfs, axis=axis)
    for column in ["level_0", "index"]:
        try:
            concat_df = concat_df.drop([column], axis=1)
        except:
            pass
    return concat_df


def toid(string: str) -> str:
    return "".join([c for c in string if c.isalnum()]).lower().strip()



import os
import pickle
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Any, Dict, Literal


EncodingType = Literal["raw", "pretrained"]
EncodingSize = Literal[32, 64, 128, 256]

ROOT_DIR = os.path.dirname(os.path.realpath(__file__))


def load_pkl_gen(gen: int):
    path = "data/encodings.pkl"
    with open(os.path.join(ROOT_DIR, path), "rb") as f:
        return pickle.load(f)[f"gen{gen}"]


def make_mlp(
    input_size: int,
    output_size: int,
    n_layers: int = 3,
    layer_norm: bool = True,
    activate_final: bool = False,
) -> nn.Module:
    layers_sizes = np.linspace(input_size, output_size, n_layers + 1).astype(int)
    layers = []
    for n, (in_feature, out_feature) in enumerate(
        zip(layers_sizes[:-1], layers_sizes[1:])
    ):
        layer = [nn.Linear(in_feature, out_feature)]
        if n < (n_layers - 1) or activate_final:
            layer.append(nn.ReLU())
            if layer_norm:
                layer.insert(1, nn.LayerNorm(out_feature))
        layers.append(nn.Sequential(*layer))
    return nn.Sequential(*layers)


class ComponentEmbedding(nn.Module):
    def __init__(
        self,
        gendata: Dict[str, Any],
        name: str,
        output_size: int,
        num_layers: int = 1,
        num_unknown: int = 1,
        use_layer_norm: bool = False,
        include_mlp: bool = True,
    ):
        super().__init__()
        self.name = name
        data = gendata[name]
        mask = data["mask"]
        self.mask = mask

        vectors = data["vectors"]
        vector_size = vectors.shape[-1]

        placeholder = np.zeros((mask.max() + 1, vector_size))
        placeholder[mask] = data["vectors"]

        self.num_unknown = num_unknown
        self.unknown = nn.Embedding(num_unknown, output_size)
        self.data = nn.Embedding.from_pretrained(torch.from_numpy(placeholder).float())

        self.include_mlp = include_mlp
        if self.include_mlp:
            self.mlp = make_mlp(
                vector_size, output_size, n_layers=num_layers, layer_norm=use_layer_norm
            )
        self.names = data["names"].tolist()

    def forward_indices(self, indices: torch.Tensor):
        return self.data((indices - self.num_unknown).clamp(min=0))

    def forward(self, indices: torch.Tensor):
        if self.include_mlp:
            return torch.where(
                (indices < self.num_unknown).unsqueeze(-1),
                self.unknown(indices.clamp(min=0, max=self.num_unknown - 1)),
                self.mlp(self.data((indices - self.num_unknown).clamp(min=0))),
            )


class PokEmb(nn.Module):
    def __init__(
        self,
        gen: int = 9,
        output_size: EncodingSize = None,
        use_layer_norm: bool = False,
        include_mlp: bool = False,
    ):
        super().__init__()

        gendata = load_pkl_gen(gen)

        self.species = ComponentEmbedding(
            gendata=gendata,
            name="species",
            num_unknown=2,
            use_layer_norm=use_layer_norm,
            include_mlp=include_mlp,
            output_size=output_size,
        )
        self.moves = ComponentEmbedding(
            gendata=gendata,
            name="moves",
            num_unknown=3,
            use_layer_norm=use_layer_norm,
            include_mlp=include_mlp,
            output_size=output_size,
        )
        self.abilities = ComponentEmbedding(
            gendata=gendata,
            name="abilities",
            num_unknown=2,
            use_layer_norm=use_layer_norm,
            include_mlp=include_mlp,
            output_size=output_size,
        )
        self.items = ComponentEmbedding(
            gendata=gendata,
            name="items",
            num_unknown=2,
            use_layer_norm=use_layer_norm,
            include_mlp=include_mlp,
            output_size=output_size,
        )

    def forward_species(self, indices: torch.Tensor) -> torch.Tensor:
        return self.species(indices)

    def forward_moves(self, indices: torch.Tensor) -> torch.Tensor:
        return self.moves(indices)

    def forward_abilities(self, indices: torch.Tensor) -> torch.Tensor:
        return self.abilities(indices)

    def forward_items(self, indices: torch.Tensor) -> torch.Tensor:
        return self.items(indices)
